@@ -1,12 +1,16 @@
+from datetime import datetime
 from UserDict import DictMixin
 from lxml import etree
+from lxml.builder import ElementMaker
+from lxml.etree import DocumentInvalid
 import os
 import re
 from copy import deepcopy
 import logging
 from pyff.utils import schema
-import pyff.xmlsec as xmlsec
+import xmlsec
 from pyff.constants import NS
+import traceback
 
 __author__ = 'leifj'
 
@@ -16,11 +20,39 @@ def _is_self_signed_err(ebuf):
             return True
     return False
 
+etree.set_default_parser(etree.XMLParser(resolve_entities=False))
+
 class MDRepository(DictMixin):
     def __init__(self):
+        """
+        A class representing a set of sets of SAML metadata.
+        """
         self.md = {}
 
-    def parse_metadata(self,fn,key=None,url=None):
+    def extensions(self,e):
+        ext = e.find("{%s}Extensions" % NS['md'])
+        if ext is None:
+            e.insert(0,etree.Element("{%s}Extensions" % NS['md']))
+            ext = e.find("{%s}Extensions" % NS['md'])
+        return ext
+
+    def annotate(self,e,category,title,message,source=None):
+        if e.tag != "{urn:oasis:names:tc:SAML:2.0:metadata}EntityDescriptor" and \
+           e.tag != "{urn:oasis:names:tc:SAML:2.0:metadata}EntitiesDescriptor":
+            raise ValueError("I can only annotate EntityDescriptor or EntitiesDescriptor elements")
+        subject = e.get('Name',e.get('entityID',None))
+        atom = ElementMaker(nsmap={'atom':'http://www.w3.org/2005/Atom'},namespace='http://www.w3.org/2005/Atom')
+        args = [atom.published("%s" % datetime.now().isoformat()),
+                atom.link(href=subject,rel="saml-metadata-subject")]
+        if source is not None:
+                args.append(atom.link(href=source,rel="saml-metadata-source"))
+        args.extend([atom.title(title),
+                     atom.category(term=category),
+                     atom.content(message,type="text/plain")])
+        self.extensions(e).insert(0,atom.entry(*args))
+
+
+    def parse_metadata(self,fn,key=None,url=None,fail_on_error=False):
         """
 Parse a piece of XML and split it up into EntityDescriptor elements. Each such element
 is stored in the MDRepository instance.
@@ -28,12 +60,29 @@ is stored in the MDRepository instance.
 :param fn: a file-like object containing SAML metadata
 :param key: a certificate (file) or a SHA1 fingerprint to use for signature verification
         """
-        logging.debug("parsing %s" % fn)
-        t = etree.parse(fn)
+        src_desc = "%s" % fn
+        if url is not None:
+            src_desc = url
+        logging.debug("parsing %s" % src_desc)
+        try:
+            t = etree.parse(fn,parser=etree.XMLParser(resolve_entities=False))
+            schema().assertValid(t)
+        except DocumentInvalid,ex:
+            logging.debug(ex.error_log)
+            raise ValueError("XML schema validation failed")
+        except Exception,ex:
+            logging.DEBUG(schema().error_log)
+            logging.error(ex)
+            if fail_on_error:
+                raise ex
+            return []
         if key is not None:
             try:
+                logging.debug("verifying signature using %s" % key)
                 xmlsec.verify(t,key)
             except Exception,ex:
+                tb = traceback.format_exc()
+                print tb
                 logging.error(ex)
                 return []
         if url is None:
@@ -42,6 +91,10 @@ is stored in the MDRepository instance.
                 url = top[0].get("Name",None)
         if url is not None:
             self[url] = t
+        # we always clean incoming ID
+        for e in t.xpath("//md:EntityDescriptor",namespaces=NS):
+            if e.attrib.has_key('ID'):
+                del e.attrib['ID']
         return t.xpath("//md:EntityDescriptor",namespaces=NS)
         #for e in t.xpath("//md:EntityDescriptor",namespaces=NS):
         #    eid = e.get('entityID')
@@ -71,7 +124,7 @@ Files ending in the specified extension are included. Directories starting with
                 for nm in files:
                     if nm.endswith(ext):
                         fn = os.path.join(top, nm)
-                        entities.extend(self.parse_metadata(fn))
+                        entities.extend(self.parse_metadata(fn,fail_on_error=True)) #local metadata is assumed to be ok
             self.md[url] = self.entity_set(entities,url)
         return self.md[url]
 
@@ -110,7 +163,7 @@ Find a (set of) EntityDescriptor element(s) based on the specified 'member' expr
         return list(set(filter(lambda x: x is not None,l)))
 
 
-    def entity_set(self,entities,name,cacheDuration=None,validUntil=None):
+    def entity_set(self,entities,name,cacheDuration=None,validUntil=None,validate=True):
         """
 Produce an EntityDescriptors set from a list of entities. Optional Name, cacheDuration and validUntil are affixed.
 
@@ -130,9 +183,21 @@ Produce an EntityDescriptors set from a list of entities. Optional Name, cacheDu
                 if ent is not None:
                     t.append(deepcopy(ent))
 
-        schema().assertValid(t)
+        if validate:
+            try:
+                schema().assertValid(t)
+            except DocumentInvalid,ex:
+                logging.debug(ex.error_log)
+                raise ValueError("XML schema validation failed")
         return t
 
+    def error_set(self,url,title,ex):
+        """
+        Creates an "error" EntitiesDescriptor - empty but for an annotation about the error that occured
+        """
+        t = etree.Element("{urn:oasis:names:tc:SAML:2.0:metadata}EntitiesDescriptor",Name=url,nsmap=NS)
+        self.annotate(t,"error",title,ex,source=url)
+
     def keys(self):
         return self.md.keys()
 
