@@ -1,3 +1,9 @@
+"""
+Package that contains the basic set of pipes - functions that can be used to put
+together a processing pipeling for pyFF.
+"""
+from lxml.etree import DocumentInvalid
+
 from pyff.utils import dumptree, schema, resource_string
 from pyff.mdrepo import NS
 from pyff.pipes import Plumbing
@@ -11,7 +17,9 @@ import hashlib
 from eventlet.green import urllib2
 from StringIO import StringIO
 from lxml import etree
-import pyff.xmlsec as xmlsec
+import xmlsec
+import base64
+from datetime import datetime
 
 __author__ = 'leifj'
 
@@ -78,7 +86,11 @@ def publish(md,t,name,args,id):
     """
     Publish the working tree.
     """
-    schema().assertValid(t)
+    try:
+        schema().assertValid(t)
+    except DocumentInvalid,ex:
+        logging.error(ex.error_log)
+        raise ValueError("XML schema validation failed")
     output_file = args.get("output",None)
     if output_file is not None:
         out = output_file
@@ -91,9 +103,9 @@ def publish(md,t,name,args,id):
 def _fetch(md,url,verify):
     logging.debug("open %s" % url)
     try:
-        return (url,urllib2.urlopen(url).read(),verify,None)
+        return url,urllib2.urlopen(url).read(),verify,None,datetime.now()
     except Exception,ex:
-        return (url,None,None,ex)
+        return url,None,None,ex,datetime.now()
 
 def _load(md,pile,args):
     """
@@ -109,7 +121,7 @@ def _load(md,pile,args):
             if len(lst) == 1:
                 url = lst[0]
                 if url.startswith("file:"):
-                    with open(url.split(":")[1]) as fd:
+                    with open(url.partition(":")[2]) as fd:
                         _load(md,pile,[line.strip() for line in fd.readlines()])
             elif len(lst) > 1:
                 url = lst[0]
@@ -124,7 +136,7 @@ def _load(md,pile,args):
 
 def remote(md,t,name,args,id):
     """
-    Load a (set of) remote URLs
+    Load a (set of) remote URLs, validate (XSD) and optionally verify signature
     """
     pool = eventlet.GreenPool()
     pile = eventlet.GreenPile(pool)
@@ -133,7 +145,8 @@ def remote(md,t,name,args,id):
 
     _load(md,pile,args)
 
-    for url,r,verify,ex in pile:
+    for url,r,verify,ex,ts_start in pile:
+        ts_end = datetime.now()
         if r is not None:
             logging.debug("url=%s: read %s bytes" % (url,len(r)))
             eids = md.parse_metadata(StringIO(r),key=verify,url=url)
@@ -151,6 +164,16 @@ def select(md,t,name,args,id):
         args = [args]
     return md.entity_set(args,id)
 
+def pick(md,t,name,args,id):
+    """
+    Select a working set of EntityDescriptor elements but don't validate it. Useful for testing.
+    """
+    if args is None:
+        args = md.keys()
+    if type(args) is str or type(args) is unicode:
+        args = [args]
+    return md.entity_set(args,id,validate=False)
+
 def sign(md,t,name,args,id):
     """
     Return a signed tree
@@ -164,8 +187,11 @@ def sign(md,t,name,args,id):
     key_file = args.get('key',None)
     cert_file = args.get('cert',None)
 
-    if key_file is None or cert_file is None:
-        raise ValueError("Empty key and cert arguments to sign pipe")
+    if key_file is None:
+        raise ValueError("Missing key argument for sign pipe")
+
+    if cert_file is None:
+        logging.info("Attempting to extract certificate from token...")
 
     xmlsec.sign(t,key_file,cert_file)
 
@@ -211,13 +237,13 @@ def xslt(md,t,name,args,id):
     stylesheet = args.pop('stylesheet',None)
     if stylesheet is not None:
         if t is None:
-            raise Exception,"Your plumbing is missing a select statement."
+            raise ValueError("Your plumbing is missing a select statement.")
         xslt = etree.fromstring(resource_string(stylesheet,"xslt"))
         transform = etree.XSLT(xslt)
         # this is to make sure the parameters are passed as xslt strings
         d = dict((k,"\'%s\'" % v) for (k,v) in args.items())
         ot = transform(t,**d)
-        t = ot.getroot()
+        t = ot #.getroot()
     return t
 
 def validate(md,t,name,args,id):
@@ -227,3 +253,47 @@ def validate(md,t,name,args,id):
     """
     if t is not None:
         schema().assertValid(t)
+
+def _subject(cert):
+    return "/".join(["%s=%s" % (c[0],c[1]) for c in cert.get_subject().get_components()])
+
+def certreport(md,t,name,args,id):
+    """
+    Generate a report of the certificates (optionally limited by expiration time) found in the selection.
+    """
+    try:
+        from OpenSSL import crypto
+    except ImportError,ex:
+        logging.error("certreport requires pyOpenSSL")
+        return t
+
+    if t is None:
+        raise ValueError("Your plumbing is missing a select statement.")
+
+    seen = {}
+    for eid in t.xpath("//md:EntityDescriptor/@entityID",namespaces=NS):
+        for cd in t.xpath("md:EntityDescriptor[@entityID='%s']//ds:X509Certificate" % eid,namespaces=NS):
+            try:
+                cert_pem = cd.text
+                cert_der = base64.b64decode(cert_pem)
+                m = hashlib.sha1()
+                m.update(cert_der)
+                fp = m.hexdigest()
+                if not seen.get(fp,False):
+                    seen[fp] = True
+                    cert = crypto.load_certificate(crypto.FILETYPE_ASN1,cert_der)
+                    et = datetime.strptime(cert.get_notAfter(),"%Y%m%d%H%M%SZ")
+                    now = datetime.now()
+                    dt = et - now
+                    if dt.total_seconds() < 0:
+                        e = cd.getparent().getparent().getparent().getparent().getparent()
+                        md.annotate(e,"certificate-error","certificate has expired","%s expired %s ago" % (_subject(cert),-dt))
+                        logging.error("%s expired %s ago" % (eid,-dt))
+                    elif dt.total_seconds() < 864000:
+                        e = cd.getparent().getparent().getparent().getparent().getparent()
+                        print e
+                        md.annotate(e,"certificate-warning","certificate about to expire","%s expires in %s" % (_subject(cert),dt))
+                        logging.warn("%s expires in %s" % (eid,dt))
+            except Exception,ex:
+                logging.error(ex)
+
